#################################
#
# Public interface
#
#################################
from ._util.casacore import __load_casa_image_block, __xds_to_casa_image
from ._util.fits import __read_fits_image
from ._util.image_factory import __make_empty_sky_image
from ._util.zarr import __xds_to_zarr, __xds_from_zarr
import warnings, time, os, logging
import numpy as np
import astropy.wcs
import xradio
from astropy import units as u
from typing import Union
import xarray as xr

warnings.filterwarnings("ignore", category=FutureWarning)


def read_image(infile: str, chunks: dict = {}, verbose: bool = False) -> xr.Dataset:
    """
    Read Image (currently only supports casacore images) or zarr to ngCASA image format
    ngCASA image spec is located at
    https://docs.google.com/spreadsheets/d/1WW0Gl6z85cJVPgtdgW4dxucurHFa06OKGjgoK8OREFA/edit#gid=1719181934
    :param infile: Path to the input CASA image
    :type infile: str, required
    :param chunks: The desired dask chunk size. Only applicable for casacore and fits images.
                   Supported optional keys are 'l', 'm', 'frequency', 'polarization',
                   and 'time'. The supported values are positive integers,
                   indicating the length of a chunk on that particular axis. If
                   a key is missing, then the associated chunk length along that axis
                   is equal to the number of pixels along that axis.
                   for zarr images, this parameter is ignored and the chunk size
                   used to store the arrays in the zarr image is used.
                   'l' represents the longitude like dimension, and 'm'
                   represents the latitude like dimension. For apeature images,
                   'u' may be used in place of 'l', and 'v' in place of 'm'.
    :type chunks: dict, required
    :param verbose: emit debugging messages? Default is False.
    :type verbose: bool, optional
    :return: xr.Dataset image that conforms to the ngCASA image spec
    :rtype: xr.Dataset
    """
    do_casa = True
    emsgs = []
    try:
        from ._util.casacore import __read_casa_image
    except Exception as e:
        emsgs.append(
            "python-casacore could not be imported, will not try to "
            f"read as casacore image: {e.args}"
        )
        do_casa = False
    if do_casa:
        try:
            return __read_casa_image(infile, chunks, verbose=verbose)
        except Exception as e:
            emsgs.append(f"image format appears not to be casacore: {e.args}")
    try:
        return __read_fits_image(infile, chunks, verbose)
    except Exception as e:
        emsgs.append(f"image format appears not to be fits {e.args}")
    try:
        return __xds_from_zarr(infile, True)
    except Exception as e:
        emsgs.append(f"image format appears not to be zarr {e.args}")
    emsgs.insert(
        0, f"Unrecognized image format. Supported types are casacore and zarr.\n"
    )
    raise RuntimeError("\n".join(emsgs))


def load_image(infile: str, block_des: dict = {}) -> xr.Dataset:
    """Load an image or portion of an image (subimage) into memory
    :param infile: Path to the input image, currently only casacore images are supported
    :type infile: str, required
    :param block_des: The description of data to return, supported keys are time,
        polarization, frequency, l (or u if apeture image), m (or v if apeture image) a
        missing key indicates to return the entire axis length for that
        dimension. Values can be non-negative integers or slices. Slicing
        behaves as numpy slicing does, that is the start pixel is included in
        the selection, and the end pixel is not. An empty dictionary (the
        default) indicates that the entire image should be returned.
    :type block_des: dict
    :return: dataset with all numpy arrays, dask arrays cannot be used in the
        implementation of this function since this function will be called in a
        delayed context.
    :rtype: xr.Dataset, all contained arrays must be numpy arrays, not dask arrays
    """
    do_casa = True
    emsgs = []

    from ._util.casacore import __read_casa_image

    try:
        from ._util.casacore import __read_casa_image
    except Exception as e:
        emsgs.append(
            "python-casacore could not be imported, will not try to "
            f"read as casacore image: {e.args}"
        )
        do_casa = False
    if do_casa:
        try:
            return __load_casa_image_block(infile, block_des)
        except Exception as e:
            emsgs.append(f"image format appears not to be casacore: {e.args}")
    """
    try:
        return __read_fits_image(infile, chunks, masks, history, verbose)
    except Exception as e:
        emsgs.append(f'image format appears not to be fits {e.args}')
    """
    try:
        return __xds_from_zarr(infile, False).isel(block_des)
    except Exception as e:
        emsgs.append(f"image format appears not to be zarr {e.args}")
    emsgs.insert(
        0, f"Unrecognized image format. Supported formats are casacore and zarr.\n"
    )
    raise RuntimeError("\n".join(emsgs))


def write_image(xds: xr.Dataset, imagename: str, out_format: str = "casa") -> None:
    """Convert xds image to CASA image
    :param xds: XDS to convert
    :type xds: xr.Dataset, required
    :param imagename: path to output CASA image
    :type imagename: str
    :param out_format: format of output image, currently only 'casa' is supported
    :type out_format: str
    """
    my_format = out_format.lower()
    if my_format == "casa":
        __xds_to_casa_image(xds, imagename)
    elif my_format == "zarr":
        __xds_to_zarr(xds, imagename)
    else:
        raise ValueError(
            f"Writing to format {out_format} is not supported. "
            'out_format must be either "casa" or "zarr".'
        )


def make_empty_sky_image(
    xds: xr.Dataset,
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    cell_size: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    pol_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
    direction_reference: str = "FK5",
    projection: str = "SIN",
    spectral_reference: str = "lsrk",
) -> xr.Dataset:
    """
    Create an image xarray.Dataset with only coordinates (no datavariables).
    The image dimensionality is either:
        l, m, time, chan, pol

    Parameters
    ----------
    xds : xarray.Dataset
        Empty dataset (dataset = xarray.Dataset()) to be modified
    phase_center : array of float, length = 2, units = rad
        Image phase center.
    image_size : array of int, length = 2, units = rad
        Number of x and y axis pixels in image.
    cell_size : array of float, length = 2, units = rad
        Cell size of x and y axis pixels in image.
    chan_coords : list or np.ndarray
        The center frequency in Hz of each image channel.
    pol_coords : list or np.ndarray
        The polarization code for each image polarization.
    time_coords : list or np.ndarray
        The time for each temporal plane in MJD.
    direction_reference : str, default = 'FK5'
    projection : str, default = 'SIN'
    spectral_reference : str, default = 'lsrk'
    Returns
    -------
    xarray.Dataset
    """
    return __make_empty_sky_image(
        xds,
        phase_center,
        image_size,
        cell_size,
        chan_coords,
        pol_coords,
        time_coords,
        direction_reference,
        projection,
        spectral_reference,
    )



import os
import re
import gdown
import shutil
import json

from astrohack._utils._tools import _remove_suffix

from prettytable import PrettyTable

FILE_ID = {
    "Antennae_South.cal.ms": "1f6a6hge0mDZVi3wUJYjRiY3Cyr9HvqZz",
    "Antennae_North.cal.ms": "1sASTyp4gr4PzWZwJr_ZHEdkqcYjF86BT",
    "demo_simulated.im": "1esOGbRMMEZXvTxQ_bdcw3PaZLg6XzHC5",
}


def check_download(name, folder, id):
    fullname = os.path.join(folder, name)
    if not os.path.exists(fullname):
        url = "https://drive.google.com/u/0/uc?id=" + id + "&export=download"
        gdown.download(url, fullname + ".zip")
        shutil.unpack_archive(filename=fullname + ".zip", extract_dir=folder)


def create_folder(path):
    if not os.path.exists(path):
        os.makedirs(path)


def build_folder_structure(dataname, resultname):
    create_folder(dataname)
    create_folder(resultname)


def download(file, folder=".", unpack=False):
    """Allows access to stakeholder and unit testing data and configuration files via gdown.

    :param file: File to download for gdirve storage. A list of the available measurement sets can be accessed via `astrohack.datasets.list_datasets()`.
    :type file: str
    :param folder: Destination folder if not the current directory, defaults to '.'
    :type folder: str, optional
    :param unpack: Unzip file, defaults to False
    :type unpack: bool, optional
    """

    # print("The google-drive download option is deprecated and will be removed soon!! Please use the dropbox option in download.")

    if file == "vla-test":
        matched = [
            (key, value)
            for key, value in FILE_ID.items()
            if re.search(
                r"^vla.+(before|after).split.+(holog|image|panel|point).*zarr$", key
            )
        ]
        files = files = list(dict(matched).keys())

    elif file == "alma-test":
        matched = [
            (key, value)
            for key, value in FILE_ID.items()
            if re.search(r"^alma.split.+(holog|image|panel|point).*zarr$", key)
        ]
        files = list(dict(matched).keys())

    else:
        files = [file]

    for file in files:
        assert (
            file in FILE_ID
        ), "File {file} not available. Available files are:".format(file=file) + str(
            FILE_ID.keys()
        )

        id = FILE_ID[file]
        create_folder(folder)

        fullname = os.path.join(folder, file)

        if os.path.exists(fullname) or os.path.exists(fullname + ".zip"):
            continue

        if unpack:
            fullname = fullname + ".zip"

        url = "https://drive.google.com/u/0/uc?id=" + id + "&export=download"
        gdown.download(url, fullname)

        # Unpack results
        if unpack:
            shutil.unpack_archive(filename=fullname, extract_dir=folder)

            # Let's clean up after ourselves
            os.remove(fullname)


def list_datasets():
    table = PrettyTable()
    table.field_names = ["Measurement Table", "Description"]
    table.align = "l"

    for key, _ in FILE_ID.items():
        basename = key.split(".")[0]
        file = "".join((basename, ".json"))
        path = os.path.dirname(__file__)

        with open(
            "{path}/data/.file_meta_data/{file}".format(path=path, file=file)
        ) as file:
            ms_info = json.load(file)

        description_string = f"""
        Observer: {ms_info['observer']}
        Project:{ms_info['project']}
        Elapsed Time: {ms_info['elapsed time']}
        Observed: {ms_info['observed']}
        SPW-ID: {ms_info['spwID']}
        Name: {ms_info['name']}
        Channels: {ms_info['channels']}
        Frame: {ms_info['frame']}
        Channel0: {ms_info['chan0']} MHz
        Channel Width: {ms_info['chan-width']} kHz
        Total Bandwidth: {ms_info['total-bandwidth']} kHz
        Center Frequency: {ms_info['center-frequency']} MHz
        Correlations: {ms_info['corrs']}
        RA: {ms_info['ra']}
        DEC: {ms_info['dec']}
        EPOCH: {ms_info['epoch']}
        Notes: {ms_info['notes']}
        """

        table.add_row([str(key), description_string])

    print(table)


def _remove_suffix(input_string, suffix):
    """
    Removes extension suffixes from file names
    Args:
        input_string: filename string
        suffix: The suffix to be removed

    Returns: the input string minus suffix

    """
    if suffix and input_string.endswith(suffix):
        return input_string[: -len(suffix)]

    return input_string



from pathlib import Path
from typing import Dict, Tuple

import xarray as xr
import numpy as np


from .msv2_msv3 import ignore_msv2_cols
from .partition_queries import make_partition_ids_by_ddi_scan
from .subtables import subt_rename_ids
from ._tables.load_main_table import load_expanded_main_table_chunk
from ._tables.read import read_generic_table, make_freq_attrs
from ._tables.read_subtables import read_delayed_pointing_table
from .._utils.partition_attrs import add_partition_attrs
from .._utils.xds_helper import make_coords


def read_spw_ddi_ant_pol(inpath: str) -> Tuple[xr.Dataset]:
    """
    Reads the four metainfo subtables needed to load data chunks into xdss.

    :param inpath: MS path (main table)
    :return: tuple with antenna, ddi, spw, and polarization setup subtables info
    """
    spw_xds = read_generic_table(
        inpath,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    )
    ddi_xds = read_generic_table(inpath, "DATA_DESCRIPTION")
    ant_xds = read_generic_table(
        inpath, "ANTENNA", rename_ids=subt_rename_ids["ANTENNA"]
    )
    pol_xds = read_generic_table(
        inpath, "POLARIZATION", rename_ids=subt_rename_ids["POLARIZATION"]
    )
    return ant_xds, ddi_xds, spw_xds, pol_xds


def load_main_chunk(
    infile: str, chunk: Dict[str, slice]
) -> Dict[Tuple[int, int], xr.Dataset]:
    """Loads a chunk of visibility data. For every DDI, a separate
    dataset is produced.
    This is very loosely equivalent to the
    partitions.read_*_partitions functions, but in a load (not lazy)
    fashion and with an implicit single partition wrt. anything but
    DDIs.
    Metainfo (sub)tables) are not loaded, and the result is one or more
    Xarray datasets. It produces one dataset per DDI found within the
    chunk slice of time/baseline.

    :param infile: MS path (main table)
    :param chunk: specification of chunk to load

    :return: dictionary of chunk datasets (keys are spw and pol_setup IDs)
    """

    chunk_dims = ["time", "baseline", "freq", "pol"]
    if not all(key in chunk_dims for key in chunk):
        raise ValueError(f"chunks dict has unknown keys. Accepted ones: {chunk_dims}")

    ant_xds, ddi_xds, spw_xds, pol_xds = read_spw_ddi_ant_pol(infile)

    # TODO: constrain this better/ properly
    data_desc_id, scan_number, state_id = make_partition_ids_by_ddi_scan(infile, False)

    all_xdss = {}
    data_desc_id = np.unique(data_desc_id)
    for ddi in data_desc_id:
        xds, part_ids, attrs = load_expanded_main_table_chunk(
            infile, ddi, chunk, ignore_msv2_cols=ignore_msv2_cols
        )

        coords = make_coords(xds, ddi, (ant_xds, ddi_xds, spw_xds, pol_xds))
        xds = xds.assign_coords(coords)
        xds = add_partition_attrs(xds, ddi, ddi_xds, part_ids, other_attrs={})

        # freq dim needs to pull its units/measure info from the SPW subtable
        spw_id = xds.attrs["partition_ids"]["spw_id"]
        xds.freq.attrs.update(make_freq_attrs(spw_xds, spw_id))
        pol_setup_id = ddi_xds.polarization_id.values[ddi]

        chunk_ddi_key = (spw_id, pol_setup_id)
        all_xdss[chunk_ddi_key] = xds

    chunk_xdss = finalize_chunks(infile, all_xdss, chunk)

    return chunk_xdss


def finalize_chunks(
    infile: str, chunks: Dict[str, xr.Dataset], chunk_spec: Dict[str, slice]
) -> Dict[Tuple[int, int], xr.Dataset]:
    """
    Adds pointing variables to a dictionary of chunk xdss. This is
    intended to be added after reading chunks from an MS main table.

    :param infile: MS path (main table)
    :param chunks: chunk xdss
    :param chunk_spec: specification of chunk to load

    :return: dictionary of chunk xdss where every xds now has pointing
    data variables
    """
    pnt_name = "POINTING"
    pnt_path = Path(infile, pnt_name)
    if "time" in chunk_spec:
        time_slice = chunk_spec["time"]
    else:
        time_slice = None
    pnt_xds = read_delayed_pointing_table(
        str(pnt_path),
        rename_ids=subt_rename_ids.get(pnt_name, None),
        time_slice=time_slice,
    )
    pnt_xds = pnt_xds.compute()

    pnt_chunks = {
        key: finalize_chunk_xds(infile, xds, pnt_xds)
        for _idx, (key, xds) in enumerate(chunks.items())
    }

    return pnt_chunks


def finalize_chunk_xds(infile: str, chunk_xds: xr.Dataset, pointing_xds) -> xr.Dataset:
    """
    Adds pointing variables to one chunk xds.

    :param infile: MS path (main table)
    :param xds_chunk: chunks xds
    :param pointing_xds: pointing (sub)table xds

    :return: chunk xds with pointing data variables interpolated form
    the pointing (sub)table
    """

    interp_pnt = pointing_xds.interp(time=chunk_xds.time, method="nearest")

    for var in interp_pnt.data_vars:
        chunk_xds[f"pointing_{var}"] = interp_pnt[var]

    return chunk_xds



import os
import xradio


def download(file, folder=".", unpack=False, source="dropbox"):
    if os.path.exists(folder) == False:
        os.mkdir(folder)

    if source == "gdrive":
        from xradio.data._google_drive import download

        download(file=file, folder=folder, unpack=unpack)

    elif source == "dropbox":
        from xradio.data._dropbox import download

        download(file=file, folder=folder)

    else:
        print("unknown source or issue found")
        pass



import logging
from typing import Dict, List, Tuple, Union

import pandas as pd
import numpy as np
import xarray as xr

from casacore import tables

from .load import load_col_chunk
from .read_main_table import get_partition_ids, redim_id_data_vars, rename_vars
from .read import add_units_measures, convert_casacore_time, extract_table_attributes
from .write import revert_time
from .table_query import open_query, open_table_ro


def load_expanded_main_table_chunk(
    infile: str,
    ddi: int,
    chunk: Dict[str, slice],
    ignore_msv2_cols: Union[list, None] = None,
) -> xr.Dataset:
    """
    Load a chunk of data from main table into memory, with expanded
    dims: (time, baseline, freq, pols)

    :param infile: Input MS path
    :param ddi: DDI to load chunk from
    :param chunk: specification of chunk to load
    :param ignore_msv2_cols: cols that should not be loaded (deprecated MSv2 or similar)

    :return: Xarray datasets with chunk of visibility data, one per DDI (spw_id, pol_setup_id) pair
    """

    taql_where = f"where DATA_DESC_ID = {ddi}"
    taql_ddi = f"select * from $mtable {taql_where}"

    with open_table_ro(infile) as mtable:
        with open_query(mtable, taql_ddi) as tb_tool:
            if tb_tool.nrows() == 0:
                return xr.Dataset()

            xds, part_ids, attrs = load_expanded_ddi_chunk(
                infile, tb_tool, taql_where, chunk, ignore_msv2_cols
            )

    return xds, part_ids, attrs


def load_expanded_ddi_chunk(
    infile: str,
    tb_tool: tables.table,
    taql_pre: str,
    chunk: Dict[str, slice],
    ignore_msv2_cols: Union[list, None] = None,
) -> xr.Dataset:
    """
    Helper function to effectively load the chunk and produce an
    xr.Dataset from a DII once the table and initial query(ies) have
    been opened successfully.

    :param infile: Input MS path
    :param tb_tool: table query contrained to one DDI and chunk time
    range
    :param taql_pre: TaQL query used for tb_tool, with some
    pre-selection of rows and columns
    :param chunk: specification of data chunk to load
    :param ignore_msv2_cols: propagated from calling funtions

    :return: An Xarray datasets with data variables as plain numpy
    arrays loaded directly from the MS columns
    """

    # read the specified chunk of data, figure out indices and lens
    utimes, times = get_chunk_times(taql_pre, chunk)
    baselines, blines = get_chunk_baselines(tb_tool, chunk)
    tidxs, bidxs, didxs, taql_where_chunk = get_chunk_data_indices(
        taql_pre, chunk, utimes, times, baselines, blines
    )

    ctlen = min(len(utimes), times[1] - times[0] + 1)
    cblen = min(len(baselines), blines[1] - blines[0] + 1)
    mvars = load_ddi_cols_chunk(
        ctlen, cblen, tidxs, bidxs, didxs, tb_tool, chunk, ignore_msv2_cols
    )

    mcoords = {
        "time": xr.DataArray(convert_casacore_time(utimes[:ctlen]), dims=["time"]),
        "baseline": xr.DataArray(np.arange(cblen), dims=["baseline"]),
    }

    # add xds global attributes
    cc_attrs = extract_table_attributes(infile)
    attrs = {"other": {"msv2": {"ctds_attrs": cc_attrs, "bad_cols": ignore_msv2_cols}}}
    # add per data var attributes
    mvars = add_units_measures(mvars, cc_attrs)
    mcoords = add_units_measures(mcoords, cc_attrs)

    mvars = rename_vars(mvars)
    mvars = redim_id_data_vars(mvars)
    xds = xr.Dataset(mvars, coords=mcoords)

    part_ids = get_partition_ids(tb_tool, taql_where_chunk)

    # needs an ~equivalent to add_partition_attrs?
    return xds, part_ids, attrs


def load_ddi_cols_chunk(
    ctlen: int,
    cblen: int,
    tidxs: np.ndarray,
    bidxs: np.ndarray,
    didxs: np.ndarray,
    tb_tool: tables.table,
    chunk: Dict[str, slice],
    ignore_msv2_cols: Union[list, None] = None,
) -> Dict[str, np.ndarray]:
    """
    For a given chunk and DDI, load all the MSv2 columns

    :param ctlen: length of the time axis/dim of the chunk
    :param cblen: length of the baseline axis of the chunk
    :param tidxs: time axis indices
    :param bidxs: baseline axis indices
    :param didxs: (effective) data indices, excluding missing baselines
    :param tb_tool: a table/TaQL query open and being used to load columns
    :param chunk: data chunk specification
    :param ignore_msv2_cols: propagated from calling funtions

    :return: columns loaded into memory as np arrays
    """
    cols = tb_tool.colnames()

    cshapes = [
        np.array(tb_tool.getcell(col, 0)).shape
        for col in cols
        if tb_tool.iscelldefined(col, 0)
    ]
    # Assumes shapes are consistent across columns - MSv2
    chan_cnt, pol_cnt = [(csh[0], csh[1]) for csh in cshapes if len(csh) == 2][0]

    dims = ["time", "baseline", "freq", "pol"]
    mvars = {}
    # loop over each column and load data
    for col in cols:
        if (col in ignore_msv2_cols + ["TIME"]) or not tb_tool.iscelldefined(col, 0):
            continue

        cdata = tb_tool.getcol(col, 0, 1)[0]
        cell_shape = cdata.shape
        if len(cell_shape) == 0:
            col_dims = dims[:2]
            mvars[col.lower()] = xr.DataArray(
                load_col_chunk(
                    tb_tool, col, (ctlen, cblen), tidxs, bidxs, didxs, None, None
                ),
                dims=col_dims,
            )

        elif col == "UVW":
            col_dims = dims[:2] + ["uvw_coords"]
            mvars[col.lower()] = xr.DataArray(
                load_col_chunk(
                    tb_tool, col, (ctlen, cblen, 3), tidxs, bidxs, didxs, None, None
                ),
                dims=col_dims,
            )

        elif len(cell_shape) == 1:
            pols, col_dims = get_col_1d_pols(cell_shape, dims, chan_cnt, pol_cnt, chunk)
            cshape = (ctlen, cblen) + (pols[1] - pols[0] + 1,)
            mvars[col.lower()] = xr.DataArray(
                load_col_chunk(tb_tool, col, cshape, tidxs, bidxs, didxs, pols, None),
                dims=col_dims,
            )

        elif len(cell_shape) == 2:
            chans, pols = get_col_2d_chans_pols(cell_shape, chan_cnt, pol_cnt, chunk)
            cshape = (ctlen, cblen) + (chans[1] - chans[0] + 1, pols[1] - pols[0] + 1)
            col_dims = dims
            mvars[col.lower()] = xr.DataArray(
                load_col_chunk(tb_tool, col, cshape, tidxs, bidxs, didxs, chans, pols),
                dims=col_dims,
            )

    return mvars


def get_chunk_times(
    taql_pre: str, chunk: Dict[str, slice]
) -> Tuple[np.ndarray, Tuple[int, int], int]:
    """
    Produces time col/axis related values for a chunk: unique times,
    start/stop times.

    :param chunk: specification of data chunk to load
    :param taql_pre: TaQL query used for tb_tool, with some pre-selection
    of rows and columns.

    :return: array of unique times + (firsr, last) time in the chunk
    """

    taql_utimes = f"select DISTINCT TIME from $mtable {taql_pre}"
    with open_query(None, taql_utimes) as query_utimes:
        utimes = np.unique(query_utimes.getcol("TIME", 0, -1))
        # add a tol around the time ranges returned by taql
        if len(utimes) < 2:
            tol = 1e-5
        else:
            tol = np.diff(utimes).min() / 4

    if "time" in chunk:
        time_slice = chunk["time"]
        if (
            type(time_slice.start) == pd.Timestamp
            and type(time_slice.stop) == pd.Timestamp
        ):
            times = (
                revert_time(time_slice.start) - tol,
                revert_time(time_slice.stop) + tol,
            )
        elif (
            int(time_slice.start) == time_slice.start
            and int(time_slice.stop) == time_slice.stop
        ):
            # could be operator.index(time_slice.start):
            nutimes = len(utimes)
            times = (
                min(nutimes, int(time_slice.start)),
                min(nutimes, int(time_slice.stop)) - 1,
            )
        else:
            raise ValueError(
                f"Invalid time type. Not a timestamp and Cannot use as"
                f" index: {time_slice.start} (type: {type(time_slice.start)})"
            )
    else:
        times = (utimes[0], utimes[-1])

    return utimes, times


def get_chunk_baselines(tb_tool, chunk) -> Tuple[np.ndarray, Tuple[int, int], int]:
    """
    Produces the basline col/axis related values for a chunk: an array
    of baselines and the start/stop baseline indices.

    :param tb_tool: table/query opened with prev selections (time)
    :param chunk: specification of data chunk to load

    :return: array of baselines + (first, last) baseline in the chunk
    """
    # main table uses time x (antenna1,antenna2)
    ant1, ant2 = tb_tool.getcol("ANTENNA1", 0, -1), tb_tool.getcol("ANTENNA2", 0, -1)
    baselines = np.array(
        [
            str(ll[0]).zfill(3) + "_" + str(ll[1]).zfill(3)
            for ll in np.unique(np.hstack([ant1[:, None], ant2[:, None]]), axis=0)
        ]
    )

    if "baseline" in chunk:
        bl_slice = chunk["baseline"]
        blines = (int(bl_slice.start), int(bl_slice.stop))
    else:
        blines = (int(baselines[0].split("_")[0]), int(baselines[-1].split("_")[0]) - 1)

    return baselines, blines


def get_chunk_data_indices(
    taql_pre: str,
    chunk: Dict[str, slice],
    utimes: np.ndarray,
    times: Tuple[int, int],
    baselines: np.ndarray,
    blines: Tuple[int, int],
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Produces indices to pass to the casacore getcol(slice) functions
    to load the chunk of data. tidxs (time), bidxs (baseline),
    didxs (effective data indices, considering present/absent
    baselines).

    Time selection is added on top of that.

    :param utimes: array of times in the chunk
    :param times: start, stop time indices
    :param baselines: array of baselines inthe chunk
    :param blines: start, stop baseline indices

    :return: indices along the time, baseline and data (time/baseline)
    axes + the full where... defined for this chunk
    """

    taql_time = f"TIME BETWEEN {utimes[times[0]]} AND {utimes[times[1]]}"
    taql_ant = f"ANTENNA1 BETWEEN {blines[0]} and {blines[1]}"
    taql_where_chunk = f"{taql_pre} AND {taql_time} AND {taql_ant}"
    taql_chunk = f"select * from $mtable {taql_where_chunk}"
    with open_query(None, taql_chunk) as query_times_ants:
        logging.debug(
            f"Opened chunk query, with {query_times_ants.nrows()} rows. Query: {taql_chunk}"
        )
        tidxs = (
            np.searchsorted(utimes, query_times_ants.getcol("TIME", 0, -1)) - times[0]
        )
        ts_ant1, ts_ant2 = (
            query_times_ants.getcol("ANTENNA1", 0, -1),
            query_times_ants.getcol("ANTENNA2", 0, -1),
        )

        ts_bases = [
            str(ll[0]).zfill(3) + "_" + str(ll[1]).zfill(3)
            for ll in np.hstack([ts_ant1[:, None], ts_ant2[:, None]])
        ]
        bidxs = np.searchsorted(baselines, ts_bases) - blines[0]

    # some antenna 2's will be out of bounds for this chunk, store rows that are in bounds
    didxs = np.where(
        (bidxs >= 0)
        & (bidxs < min(blines[1] - blines[0] + 1, len(baselines) - blines[0]))
    )[0]

    return tidxs, bidxs, didxs, taql_where_chunk


def get_col_1d_pols(
    cell_shape: Tuple[int],
    dims: List[str],
    chan_cnt: int,
    pol_cnt: int,
    chunk: Dict[str, slice],
) -> Tuple[Tuple[int, int], List[str]]:
    """
    For a column with 1d array values, calculate the start/stop
    indices for the last dimension (either pol or freq).
    It also produces the adequate dimension names.

    :param cell_shape: shape of the column
    :param dims: full list of dataset dimensions
    :param chan_cnt: number of channels
    :param pol_cnt: number of pols
    :param chunk: data chunk specification

    :return: first and last pol/freq index of the chunk, and its
    dimension name
    """
    if cell_shape == chan_cnt:
        # chan/freq
        col_dims = dims[:2] + ["freq"]
        if "freq" in chunk:
            pols = (
                min(chan_cnt, chunk["freq"].start),
                min(chan_cnt, chunk["freq"].stop) - 1,
            )
        else:
            pols = (0, cell_shape[0])
    else:
        # pol
        col_dims = dims[:2] + ["pol"]
        if "pol" in chunk:
            pols = (
                min(pol_cnt, chunk["pol"].start),
                min(pol_cnt, chunk["pol"].stop) - 1,
            )
        else:
            pols = (0, cell_shape[0])

    return pols, col_dims


def get_col_2d_chans_pols(
    cell_shape: Tuple[int],
    chan_cnt: int,
    pol_cnt: int,
    chunk: Dict[str, slice],
) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    """
    For a column with 2d array values (FLAG, DATA, WEIGHT_SPECTRUM,
    etc., calculate the the start/stop indices for the last two
    dimensions of the chunk (freq and pol).
    The dimension names can be assumed to be the full list of dims in
    visibilities (time, baseline, freq, pol).

    :param cell_shape: shape of the column
    :param chan_cnt: number of channels
    :param pol_cnt: number of pols
    :param chunk: data chunk specification

    :return: first and last index for freq (channel) and pol axes of
    the chunk
    """
    if "freq" in chunk:
        chans = (
            min(chan_cnt, chunk["freq"].start),
            min(chan_cnt, chunk["freq"].stop) - 1,
        )
    else:
        chans = (0, cell_shape[0])

    if "pol" in chunk:
        pols = (
            min(pol_cnt, chunk["pol"].start),
            min(pol_cnt, chunk["pol"].stop) - 1,
        )
    else:
        pols = (0, cell_shape[1])

    return chans, pols



import logging
import os
from typing import Union
import xarray as xr
from ._fits.xds_from_fits import __fits_image_to_xds


def __read_fits_image(infile: str, chunks: dict, verbose: bool = False) -> xr.Dataset:
    img_full_path = os.path.expanduser(infile)
    xds = __fits_image_to_xds(img_full_path, chunks, verbose)
    return xds
